package edg.graph;

import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import edg.LDASTNodeInfo;
import edg.constraint.Constraints;
import edg.constraint.GrammarConstraint;
import edg.graphlib.Arrow;
import edg.graphlib.Graph;
import edg.graphlib.Vertex;
import edg.slicing.SlicingCriterion;
import edg.traverser.EDGTraverser;

public class EDG extends LAST
{
	public EDG()
	{
		
	}
	public EDG(LAST last)
	{
		this.graph = last.graph; 
	}
	/*****************/
	/***** Nodes *****/
	/*****************/
	public Node getNode(SlicingCriterion sc)
	{
		if (sc == null)
			return null;

		final String scArchive = sc.getArchive();
		final int scLine = sc.getLine();
		final String scName = sc.getName();
		final List<Node> nodes = this.findNodesByData(null, new Comparator<NodeInfo>() {
			public int compare(NodeInfo o1, NodeInfo o2)
			{
				final LDASTNodeInfo ldNodeInfo = o2.getInfo();
				if (ldNodeInfo == null)
					return -1;
				if (scLine != ldNodeInfo.getLine())
					return -1;
				if (!scName.equals(o2.getName()))
					return -1;
				if (!scArchive.equals(ldNodeInfo.getArchive()))
					return -1;
				return 0;
			}
		});
		final int scOccurrence = sc.getOccurrence();
		if (nodes.isEmpty())
			return null;
		final Node node = nodes.get(scOccurrence - 1);
		return EDGTraverser.getResult(node);
	}

	/*****************/
	/**** Grammar ****/
	/*****************/
	private final Grammar grammar = new Grammar();

	public Grammar getGrammar()
	{
		return this.grammar;
	}
	public List<Constraints> getProductions(GrammarConstraint grammarConstraint)
	{
		return this.grammar.getProductions(grammarConstraint);
	}

	public void addProduction(GrammarConstraint grammarConstraint, Constraints production)
	{
		this.grammar.addProduction(grammarConstraint, production);
	}
	
	/**********************************/
	/**** Generation Time Measurer ****/
	/**********************************/
	private final GraphGeneratorTimer ggt = new GraphGeneratorTimer();
	
	public void setStructureTime(double time) {	this.ggt.setStructureTime(time); }
	public void setControlFlowTime(double time) {	this.ggt.setControlFlowTime(time); }
	public void setControlTime(double time) {	this.ggt.setControlTime(time);	}
	public void setInterproceduralTime(double time) {	this.ggt.setInterproceduralTime(time); }
	public void setFlowTime(double time) {	this.ggt.setFlowTime(time); }
	public void setValueTime(double time) {	this.ggt.setValueTime(time); }
	public void setSummaryTime(double time) {	this.ggt.setSummaryTime(time); }
	public void setExceptionTime(double time) {	this.ggt.setExceptionTime(time); }
	
	public GraphGeneratorTimer getGenerationTime()
	{
		return this.ggt;
	}
	
	public static class GraphGeneratorTimer
	{	
		private double structureTime;
		private double controlFlowTime;
		private double controlTime;
		private double interproceduralTime;
		private double flowTime;
		private double valueTime;
		private double summaryTime;
		private double exceptionTime;
		
		public GraphGeneratorTimer()
		{
			
		}
		
		public void setStructureTime(double time) {	this.structureTime = time;	}
		public void setControlFlowTime(double time) {	this.controlFlowTime = time;	}
		public void setControlTime(double time) {	this.controlTime = time;	}
		public void setInterproceduralTime(double time) {	this.interproceduralTime = time;	}
		public void setFlowTime(double time) {	this.flowTime = time;	}
		public void setValueTime(double time) {	this.valueTime = time;	}
		public void setSummaryTime(double time) {	this.summaryTime = time;	}
		public void setExceptionTime(double time) {	this.exceptionTime = time;	}
		
		public double getGenerationEDGTime()
		{
			return structureTime + controlFlowTime + controlTime + interproceduralTime + flowTime + 
					 valueTime + summaryTime + exceptionTime;
		}
		public double getGenerationSDGTime(EDG edg)
		{
			final int SDGNodeCount = countSDGNodes(edg);
			final int EDGNodeCount = edg.getNodes().size();
			final double structureSDGTime = structureTime / EDGNodeCount * SDGNodeCount;
			final double controlFlowSDGTime = controlFlowTime / EDGNodeCount * SDGNodeCount;
			final double flowSDGTime = flowTime - ((SDGNodeCount * flowTime)/(2 * EDGNodeCount));
			
			return structureSDGTime + controlFlowSDGTime + controlTime + 
					interproceduralTime + flowSDGTime + summaryTime;
		}
		
		private int countSDGNodes(EDG edg)
		{
			final List<Integer> sdgNodeNumbers = new LinkedList<Integer>();
			final List<Node> edgNodes = edg.getNodes();
			for (Node node : edgNodes)
			{
				final int sdgId = node.getData().getSDGId();
				if (!sdgNodeNumbers.contains(sdgId))
					sdgNodeNumbers.add(sdgId);
			}
			return sdgNodeNumbers.size();
		}
	}
}